--- old/usr/src/uts/common/fs/zfs/dmu_send.c	2012-07-30 16:07:08.855607172 +0400
+++ new/usr/src/uts/common/fs/zfs/dmu_send.c	2012-07-30 16:07:08.291607158 +0400
@@ -850,12 +850,6 @@
 			return (EINVAL);
 		}
 
-		/* must not have an incremental recv already in progress */
-		if (!mutex_tryenter(&ds->ds_recvlock)) {
-			dsl_dataset_rele(ds, dmu_recv_tag);
-			return (EBUSY);
-		}
-
 		/* tmp clone name is: tofs/%tosnap" */
 		(void) snprintf(rbsa.clonelastname, sizeof (rbsa.clonelastname),
 		    "%%%s", tosnap);
@@ -863,12 +857,12 @@
 		err = dsl_sync_task_do(ds->ds_dir->dd_pool,
 		    recv_existing_check, recv_existing_sync, ds, &rbsa, 5);
 		if (err) {
-			mutex_exit(&ds->ds_recvlock);
 			dsl_dataset_rele(ds, dmu_recv_tag);
 			return (err);
 		}
-		drc->drc_logical_ds = ds;
+		drc->drc_logical_dsobj = ds->ds_object;
 		drc->drc_real_ds = rbsa.ds;
+		dsl_dataset_rele(ds, dmu_recv_tag);
 	} else if (err == ENOENT) {
 		/* target fs does not exist; must be a full backup or clone */
 		char *cp;
@@ -898,7 +892,7 @@
 		dsl_dataset_rele(ds, FTAG);
 		if (err)
 			return (err);
-		drc->drc_logical_ds = drc->drc_real_ds = rbsa.ds;
+		drc->drc_real_ds = rbsa.ds;
 		drc->drc_newfs = B_TRUE;
 	}
 
@@ -1528,10 +1522,6 @@
 
 		(void) dsl_dataset_destroy(drc->drc_real_ds, dmu_recv_tag,
 		    B_FALSE);
-		if (drc->drc_real_ds != drc->drc_logical_ds) {
-			mutex_exit(&drc->drc_logical_ds->ds_recvlock);
-			dsl_dataset_rele(drc->drc_logical_ds, dmu_recv_tag);
-		}
 	}
 
 	kmem_free(ra.buf, ra.bufsize);
@@ -1600,8 +1590,20 @@
 dmu_recv_existing_end(dmu_recv_cookie_t *drc)
 {
 	struct recvendsyncarg resa;
-	dsl_dataset_t *ds = drc->drc_logical_ds;
+	dsl_dataset_t *ds;
 	int err, myerr;
+	dsl_pool_t *dp = drc->drc_real_ds->ds_dir->dd_pool;
+
+	rw_enter(&dp->dp_config_rwlock, RW_READER);
+	err = dsl_dataset_own_obj(dp, drc->drc_logical_dsobj, FALSE,
+	    dmu_recv_tag, &ds);
+	rw_exit(&dp->dp_config_rwlock);
+
+	if (err) {
+		(void) dsl_dataset_destroy(drc->drc_real_ds, dmu_recv_tag,
+		    B_FALSE);
+		return (EBUSY);
+	}
 
 	/*
 	 * XXX hack; seems the ds is still dirty and dsl_pool_zil_clean()
@@ -1610,18 +1612,9 @@
 	 */
 	txg_wait_synced(ds->ds_dir->dd_pool, 0);
 
-	if (dsl_dataset_tryown(ds, FALSE, dmu_recv_tag)) {
-		err = dsl_dataset_clone_swap(drc->drc_real_ds, ds,
-		    drc->drc_force);
-		if (err)
-			goto out;
-	} else {
-		mutex_exit(&ds->ds_recvlock);
-		dsl_dataset_rele(ds, dmu_recv_tag);
-		(void) dsl_dataset_destroy(drc->drc_real_ds, dmu_recv_tag,
-		    B_FALSE);
-		return (EBUSY);
-	}
+	err = dsl_dataset_clone_swap(drc->drc_real_ds, ds, drc->drc_force);
+	if (err)
+		goto out;
 
 	resa.creation_time = drc->drc_drrb->drr_creation_time;
 	resa.toguid = drc->drc_drrb->drr_toguid;
@@ -1635,7 +1628,6 @@
 	}
 
 out:
-	mutex_exit(&ds->ds_recvlock);
 	if (err == 0 && drc->drc_guid_to_ds_map != NULL)
 		(void) add_ds_to_guidmap(drc->drc_guid_to_ds_map, ds);
 	dsl_dataset_disown(ds, dmu_recv_tag);
@@ -1648,7 +1640,7 @@
 dmu_recv_new_end(dmu_recv_cookie_t *drc)
 {
 	struct recvendsyncarg resa;
-	dsl_dataset_t *ds = drc->drc_logical_ds;
+	dsl_dataset_t *ds = drc->drc_real_ds;
 	int err;
 
 	/*
@@ -1679,7 +1671,7 @@
 int
 dmu_recv_end(dmu_recv_cookie_t *drc)
 {
-	if (drc->drc_logical_ds != drc->drc_real_ds)
+	if (!drc->drc_newfs)
 		return (dmu_recv_existing_end(drc));
 	else
 		return (dmu_recv_new_end(drc));
